/*
 * Author: Giam Teck Choon <choon@choon.net>
 *
 * Description: mailfront-imapfront-auth-dovecot is a program to modify the
 * MAIL environment variable according to dovecot imap requirements in order
 * to access users mail storage.
 * 
 * Dovecot uses colons as internal field separators; however, since v1.2rc6, it
 * allows colons occurrences via escaping (as : -> ::). Before passing such
 * mailbox paths via MAIL environment variable to dovecot imap, this wrapper
 * performs such escaping.
 * 
 * From original MAIL environment variable:
 * 
 * /home/USERNAME/MAILDIR/mail:box:user
 * 
 * To:
 * 
 * maildir:/home/USERNAME/MAILDIR/mail::box::user
 *
 * Or:
 *
 * mbox:/home/USERNAME/MAILDIR/mail::box::user
 * 
 * More information can be read at:
 * 
 * http://wiki.dovecot.org/MailLocation
 * http://wiki2.dovecot.org/MailLocation
 * 
 * References:
 * 
 * http://lists.untroubled.org/?list=bgware&cmd=showthread&month=201011&threadid=elpfjlnilcganfkokehg
 * http://mij.oltrelinux.com/net/dovecot-qmail-vmailmgr/
 * http://wiki.dovecot.org/HowTo/VMailMgr
 *
 * Copyright: This package is Copyright(C) 2010 Giam Teck Choon or CHOON.NET, and may be copied
 *            according to the GNU GENERAL PUBLIC LICENSE (GPL) Version 2 or a later version. A
 *            copy of this license is included with this package. This package comes with no
 *            warranty of any kind.
 *
 * Disclaimer: Use this program at your own risk! I am not responsible for any damages caused.
 *             You have been warned!
 *
 * Code reference:
 * http://mij.oltrelinux.com/net/dovecot-qmail-vmailmgr/checkvpw-dovecot-wrapper.c
 * MANY THANKS TO mij@bitchx.it
 */

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <regex.h>
#include <sys/stat.h>

#define PROGNAME                "mailfront-imapfront-auth-dovecot"
/* what maximum length we can expect in MAIL environment variable */
#define MAXLEN                  600

/* replace all ":" with "::" in string. Return if a replacement occurred. */
int escape_colon(char *string, size_t buflen) {
    int counter, numcolons;

    /* any ':' in original string? */
    for (numcolons = counter = 0; string[counter] != '\0'; counter++) {
        if (string[counter] == ':') numcolons++;
    }
    if (numcolons > 0) {
        char *tmpstr;
        size_t newlen;
        int j; /* index for iteration over second string */

        /* copy string -> tmpstr, duplicating all ':' -> '::' */
        newlen = strlen(string) + numcolons + 1;
        if (newlen > buflen) {
	    return 111;
        }
        tmpstr = (char *)malloc(newlen);
        /* copy everything, but put "::" when ":" is found */
        for (j = counter = 0; string[counter] != '\0'; counter++, j++) {
            tmpstr[j] = string[counter];
            if (string[counter] == ':') {
                /* add another ':' */
                j++;
                tmpstr[j] = ':';
            }
        }
        tmpstr[j] = '\0';

        /* now copy over to original string */
        strcpy(string, tmpstr);
        free(tmpstr);

        return 1;
    }

    return 0;
}

/* Below function copied from http://n.h7a.org/hacks/qmail-badremotehost.patch */
/* Another is qregex.c/qregex.h aka qmail-regex but I don't think I need it here. :p */
int is_regex_matched(char *regex, char *text) {
    int rc;
    regex_t reg;

    rc = regcomp(&reg,regex,REG_EXTENDED|REG_ICASE);
    if (!rc) rc = regexec(&reg,text,0,0,0);
    regfree(&reg);

    return rc ? 0 : 1;
}

int main(int argc, char *argv[]) {
    char buffer[MAXLEN];
    unsigned long pid;
    const char *ip;
    const char *tmp;
    char mailenv[MAXLEN+10];
    int debug = 0;
    extern char **environ;
    char *cmd[50];
    char mail_location[MAXLEN+30] = "mail_location=";
    int skip = 0;
    int c = 0;
    struct stat statinfo;

    if (argc < 2) {
	fprintf(stderr, "usage: %s program [arguments]\n", PROGNAME);
	exit(111);
    }

    pid = (unsigned long)getpid();

    /* get IP from environment variable TCPREMOTEIP */
    if (getenv("IMAPFRONTAUTH_DOVECOT_DEBUG") != 0) {
	debug = 1;
    }

    /* get IP from environment variable TCPREMOTEIP */
    if (getenv("TCPREMOTEIP") != 0) {
	ip = getenv("TCPREMOTEIP");
    } else {
	ip = "";
    }

    /* check for environment variable MAIL */
    tmp = getenv("MAIL");
    if (strlen(tmp) > 0) {
	if (strlen(tmp) > MAXLEN) {
	    if (debug > 0) {
		fprintf(stderr, "%s: %s pid %lu: MAIL environment variable length is greater than %d!\n",PROGNAME, ip, pid, MAXLEN);
	    }
	    exit(111);
	} else {
	    /* if MAIL starts with maildir: or mbox: then we skip */
	    if (is_regex_matched("^maildir:", (char *) tmp)) {
		if (debug > 0) {
		    fprintf(stderr, "%s: %s pid %lu: MAIL environment starting string match: '%s'\n", PROGNAME, ip, pid, "maildir:");
		}
		skip++;
	    }
	    if (is_regex_matched("^box:", (char *) tmp)) {
		if (debug > 0) {
		    fprintf(stderr, "%s: %s pid %lu: MAIL environment starting string match: '%s'\n", PROGNAME, ip, pid, "mbox:");
		}
		skip++;
	    }
	    if (skip == 0) {
		/* escape possible ':' characters before finally exporting */
		strcpy(buffer, tmp);
		if (escape_colon(buffer, MAXLEN) == 111) {
		    if (debug > 0) {
			fprintf(stderr, "%s: %s pid %lu: buffer too short to accommodate the resulting MAIL environment\n", PROGNAME, ip, pid);
		    }
		    exit(111);
		} else {
		    /* stat whether MAIL is a directoy for maildir or otherwise mbox is assumed */
		    if (stat(tmp, &statinfo) == 0) {
			if (S_ISDIR(statinfo.st_mode)) {
			    strcat(mailenv, "maildir:");
			    if (debug > 0) {
				fprintf(stderr, "%s: %s pid %lu: MAIL is a %s\n", PROGNAME, ip, pid, mailenv);
			    }
			} else {
			    strcat(mailenv, "mbox:");
			    if (debug > 0) {
				fprintf(stderr, "%s: %s pid %lu: MAIL is a %s\n", PROGNAME, ip, pid, mailenv);
			    }
			}
		    } else {
			strcat(mailenv, "mbox:");
			if (debug > 0) {
			    fprintf(stderr, "%s: %s pid %lu: MAIL is a %s\n", PROGNAME, ip, pid, mailenv);
			}
		    }
		    strcat(mailenv, buffer);
		    setenv("MAIL", mailenv, 1);
		    if (debug > 0) {
			fprintf(stderr, "%s: %s pid %lu: MAIL environment is %s\n", PROGNAME, ip, pid, mailenv);
		    }
		}
	    } else {
		strcpy(mailenv, tmp);
		if (debug > 0) {
		    fprintf(stderr, "%s: %s pid %lu: MAIL environment modification skipped: %s\n", PROGNAME, ip, pid, mailenv);
		}
	    }
	    /* If $DOVECOT_V2 is set */
	    if (getenv("DOVECOT_V2") != 0) {
		if (debug > 0) {
		    fprintf(stderr, "%s: %s pid %lu: dovecot imap command options: %s%s\n", PROGNAME, ip, pid, "-O -o mail_location=", mailenv);
		    c = 0;
		    while(environ[c]) {
			fprintf(stderr, "%s: %s pid %lu: environ[%d]: %s\n",PROGNAME, ip, pid, c, environ[c]);
			c++;
		    }
		}
		/* append custom dovecot v2 imap options */
		c = 0;
		while (argv[c]) {
		    cmd[c] = argv[c];
		    if (debug > 0) {
			fprintf(stderr, "%s: %s pid %lu: cmd[%d]: %s\n",PROGNAME, ip, pid, c, cmd[c]);
		    }
		    c++;
		}
		cmd[c] = "-O";
		if (debug > 0) {
		    fprintf(stderr, "%s: %s pid %lu: cmd[%d]: %s\n",PROGNAME, ip, pid, c, cmd[c]);
		}
		c++;
		cmd[c] = "-o";
		if (debug > 0) {
		    fprintf(stderr, "%s: %s pid %lu: cmd[%d]: %s\n",PROGNAME, ip, pid, c, cmd[c]);
		}
		c++;
		strcat(mail_location, mailenv);
		cmd[c] = mail_location;
		if (debug > 0) {
		    fprintf(stderr, "%s: %s pid %lu: cmd[%d]: %s\n",PROGNAME, ip, pid, c, cmd[c]);
		}
		c++;
		cmd[c] = (char *)0; /* null terminated */
		if (debug > 0) {
		    c = 0;
		    while (argv[c]) {
			fprintf(stderr, "%s: %s pid %lu: argv[%d]: %s\n",PROGNAME, ip, pid, c, argv[c]);
			c++;
		    }
		    c = 0;
		    while (cmd[c]) {
			fprintf(stderr, "%s: %s pid %lu: cmd[%d]: %s\n",PROGNAME, ip, pid, c, cmd[c]);
			c++;
		    }
		}
		execvp(argv[1], cmd+1);
	    } else {
		execvp(argv[1], argv+1);
	    }
	    fprintf(stderr, "%s: %s pid %lu: unable to run %s: %s\n",PROGNAME, ip, pid, argv[1], strerror(errno));
	    exit(111);
	}
    } else {
	fprintf(stderr, "%s: %s pid %lu: MAIL environment variable not found!\n",PROGNAME, ip, pid);
	printf("MAIL environment variable is required!\r\n");
	exit(111);
    }
}
